package org.greenstone.tabletop;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.nio.charset.StandardCharsets;
import java.net.URL;
import java.net.MalformedURLException;
    
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.io.FileUtils;

import javax.servlet.Servlet;
import javax.servlet.ServletConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.json.JSONObject;

//import org.eclipse.jetty.server.Server;

import org.greenstone.tabletop.action.BaseAction;
import org.greenstone.tabletop.action.CheckExistsAction;


/**
 * Servlet implementation class to act as a bridge between 
 * Greenstone3 and the XML format the Touch Tabletop software expects
 */
public class ConnectProxyServlet extends HttpServlet 
{
    private static final long serialVersionUID = 1L;
    protected static Logger logger = Logger.getLogger(ConnectProxyServlet.class.getName());
    
    protected static CheckExistsAction check_exists_ = null;
    
    protected static ArrayList<BaseAction> action_list_ = null;

    public ConnectProxyServlet() {
    }

    protected static Map<String,List<String>> convertParamMapArrayToMapList(Map<String,String []> map_array)
    {
	Map<String,List<String>> map_list = new HashMap<String,List<String>>();
		
	for (String key: map_array.keySet()) {
	    List<String> val_list = Arrays.asList(map_array.get(key));
			
	    map_list.put(key, val_list);
	}
	
	return map_list;
    }

    /*
    // keep around, could be useful if needing to convert Map data in the other direction
    protected static Map<String,String[]> convertParamMapListToMapArray(Map<String,List<String>> map_list)
    {
    Map<String,String[]> map_array = new HashMap<String,String[]>();
		
    for (String key: map_list.keySet()) {
    String [] val_array = (String[]) map_list.get(key).toArray();
    map_array.put(key, val_array );
    }
	
    return map_array;
    }
    */
	
    
    /**
     * @see Servlet#init(ServletConfig)
     */
    public void init(ServletConfig config) throws ServletException {
	super.init(config);


	String mongo_host = config.getInitParameter("mongodbHost");

	ServletContext context = getServletContext();

	if (check_exists_ == null) {	
	    check_exists_ = new CheckExistsAction(context,config);
	}

		
	if (action_list_ == null) {
	    action_list_ = new ArrayList<BaseAction>();
	    action_list_.add(check_exists_);
	}		
    }	


    protected void displayUsage(HttpServletResponse http_servlet_response)
    {
	try {
	    PrintWriter pw = http_servlet_response.getWriter();

	    pw.append("Usage:\n");

	    for (BaseAction action: action_list_) {
			    
		pw.append("  action=" + action.getHandle() + "\n");
		String[] mess = action.getDescription();
		for (String sm: mess) {
		    pw.append("    " + sm + "\n");
		}
		pw.append("\n");
	    }
	    pw.close();
	}
	catch (Exception e) {
	    e.printStackTrace();
	}

    }

    public String escapeHTML(String s)
    {
	StringBuilder out = new StringBuilder(Math.max(16, s.length()));
	for (int i = 0; i < s.length(); i++) {
	    char c = s.charAt(i);
	    if (c == '"' || c == '<' || c == '>' || c == '&') {
		out.append("&#");
		out.append((int) c);
		out.append(';');
	    } else {
		out.append(c);
	    }
	}
	return out.toString();
    }
    
    protected String convertGS3UrlToFile(String gs3_url)
    {
	String gs3_file = gs3_url
	    .replace("/greenstone3/library","C:/Table/Projects/greenstone3-svn/web")
	    .replaceAll("/","\\\\");

	return gs3_file;
    }
    
	
    protected String convertGS3Urls(String xml_text)
    {
	Pattern gsdlurl_pattern = Pattern.compile("(<GS3URL>)(.*?)(</GS3URL>)");
	Matcher gsdlurl_matcher = gsdlurl_pattern.matcher(xml_text);

	StringBuilder mapped_output_builder = new StringBuilder();

	int last_index = 0;

	while (gsdlurl_matcher.find()) {
	    mapped_output_builder.append(xml_text, last_index, gsdlurl_matcher.start());
	    mapped_output_builder.append(convertGS3UrlToFile(gsdlurl_matcher.group(2)));
	    
	    last_index = gsdlurl_matcher.end();
	}
	if (last_index < xml_text.length()) {
	    mapped_output_builder.append(xml_text, last_index, xml_text.length());
	}

	String mapped_output = mapped_output_builder.toString();
	
	return mapped_output;
    }

    
	
    protected void doTblPeople(HttpServletRequest request, HttpServletResponse response)
	throws MalformedURLException, IOException
    {
	URL dl_url = new URL("http://localhost:8383/greenstone3/library/collection/computing-50-tabletop/browse/CL1?sa=tabletop&excerptid=classifiers");

	
	//File srcFile = new File(__dirname + "/" + export_filename);
	//OutputStream os = response.getOutputStream();
        
	InputStreamReader isr = new InputStreamReader(dl_url.openStream(),StandardCharsets.UTF_8);
	BufferedReader in = new BufferedReader(isr);

	response.setHeader("Content-type", "text/xml");
	response.setCharacterEncoding("UTF-8");
	
	PrintWriter pw = response.getWriter();
	pw.append("<?xml version=\"1.0\" encoding=\"utf-8\"?>\n");
		
        String input_line;
        while ((input_line = in.readLine()) != null) {
	    //logger.info("input_line = '" + input_line + "'");
	    String converted_input_line1 = convertGS3Urls(input_line);
	    String converted_input_line2 = converted_input_line1.replaceAll("<query id=\"classifiers\">","<query>");
	    String converted_input_line3 = converted_input_line2.replaceAll("<c/>","<c></c>");
	    String converted_input_line_final = converted_input_line3.replaceAll("<c />","<c></c>");

	    if (!converted_input_line_final.matches("^\\s*$")) {
		//logger.info("converted_line = " + converted_input_line_final);
		pw.append(converted_input_line_final + "\n");
	    }
	    
	}
    }
    
    protected void doGet(HttpServletRequest request, HttpServletResponse response)
	throws ServletException, IOException 
    {
	//logger.info("Handling path quest: request =" + request.toString());

    	Map<String, String[]> param_map_array = request.getParameterMap();

	String[] query_args = param_map_array.get("query");

	if ((query_args != null) && (query_args.length>0)) {
	    String query_arg = query_args[query_args.length-1];

	    String[] export_table_names = new String[] {
		"Tbl_Acknowledgements",
		"Tbl_CoolStuff",
		"Tbl_EventType",
		"Tbl_MajorEvents",
		"Tbl_Maps",
		"Tbl_People",
		"Tbl_Per_Events",
		"Tbl_Pictures",
		"Tbl_Sponsors",
		"Tbl_Timeline",
		"Tbl_Videos",
		"Team"
	    };
	
	    String export_filename = null;
	    
	    for (String table_name: export_table_names) {	    
		if (query_arg.indexOf(table_name)>=0) {
		    logger.info("ConnectProxy matching on table-name: " + table_name);
		    export_filename = table_name + ".xml";
		    break;
		}
	    }

	    //final String __dirname = ".";
	    final String __dirname = "C:/Temp";
	    
	    if (export_filename != null) {

		if (export_filename.equals("Tbl_People.xml")) {
		    doTblPeople(request,response);
		}
		else {

		    String waikato_export_filename = "Waikato_" + export_filename;
		    File waikato_src_file = new File(__dirname + "/" + waikato_export_filename);

		    File srcFile = null;
		    if (waikato_src_file.exists()) {
			logger.info("Detected Waikato version of: " + export_filename);
			srcFile = waikato_src_file;
		    }
		    else {
			srcFile = new File(__dirname + "/" + export_filename);
		    }
		    
		    response.setHeader("Content-type", "text/xml");
		    response.setCharacterEncoding("UTF-8");
		    
		    FileUtils.copyFile(srcFile, response.getOutputStream());
		}
	    }
	    else {
		logger.log(Level.SEVERE, "Requessted URL did not match one of the predefined table names");
		logger.log(Level.SEVERE, Arrays.toString(export_table_names));
		response.sendError(HttpServletResponse.SC_NOT_FOUND);
	    }
	}
	else {
	    try {
		PrintWriter pw = response.getWriter();
		
		response.setHeader("Content-type", "text/xml");	    
		pw.append("<?xml version=\"1.0\" encoding=\"utf-8\"?>\n<ping />");
		logger.info("Responding with 'ping'");

		pw.close();
	    }
	    catch (Exception e) {
		e.printStackTrace();
		response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
	    }
	}
    }
	
    protected void doGet_ACTION_BASED(HttpServletRequest request, HttpServletResponse response)
	throws ServletException, IOException 
    {
	Map<String, String[]> param_map_array = request.getParameterMap();
	Map<String, List<String>> param_map_list = convertParamMapArrayToMapList(param_map_array);
		
	boolean action_match = false;
	String action_handle = BaseAction.getParameter(param_map_list,"action");

	for (BaseAction action: action_list_) {
	    if (action.getHandle().equals(action_handle)) {
		action_match = true;
				
		action.doAction(param_map_list,response);
		break;
	    }
	}

	if (!action_match) {
	    // No action given => generate usage statement
	    response.setContentType("text/plain");
	    displayUsage(response);
	}
    }


    /**
     * @see HttpServlet#doPost(HttpServletRequest request, HttpServletResponse response)
     */
    protected void doPost(HttpServletRequest request, HttpServletResponse response)
	throws ServletException, IOException {
	doGet(request, response);
    }


}
